# -*- coding: utf-8 -*-

# Copyright (C) 2012-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
# Copyright (C) 2012-2013 by Oleksandr Shneyder <oleksandr.shneyder@obviously-nice.de>
# Copyright (C) 2012-2013 by Heinz-Markus Graesing <heinz-m.graesing@obviously-nice.de>
#
# X2Go Session Broker is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# X2Go Session Broker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import subprocess
import paramiko
import cStringIO
import time
import threading
import socket

import x2gobroker._paramiko
x2gobroker._paramiko.monkey_patch_paramiko()

# X2Go Broker modules
import x2gobroker.defaults
import x2gobroker.x2gobroker_exceptions
from x2gobroker.loggers import logger_broker, logger_error


class delayed_execution(threading.Thread):

    def __init__(self, agent_func, delay=0, **kwargs):
        threading.Thread.__init__(self)
        self.agent_func = agent_func
        self.delay = delay
        self.kwargs = kwargs
        self.daemon = True
        self.start()

    def run(self):
        i = 0
        while i < self.delay:
            time.sleep(1)
            i += 1
        self.agent_func(**self.kwargs)


def call_local_broker_agent(username, mode, cmdline_args=[]):
    """\
    Launch X2Go Broker Agent locally and process its output.

    @param username: run the broker agent for this user
    @type username: C{unicode}
    @param mode: execution mode of the broker (listsessions, getservers, etc.)
    @type mode: C{unicode}

    """
    cmd_line = [
        '{x2gobroker_agent_binary}'.format(x2gobroker_agent_binary=x2gobroker.defaults.X2GOBROKER_AGENT_CMD),
        '{username}'.format(username=username),
        '{mode}'.format(mode=mode),
    ]

    for cmdline_arg in cmdline_args:
        cmd_line.append('\'{arg}\''.format(arg=cmdline_arg))

    logger_broker.debug('Executing agent command locally: {cmd}'.format(cmd=" ".join(cmd_line)))
    agent_process = subprocess.Popen(cmd_line,
                                     stdin=None,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT,
                                     shell=False,
    )

    result = agent_process.stdout.read().split('\n')
    if result[0].startswith('OK'):
        return [ r for r in result[1:] if r ]

    raise x2gobroker.x2gobroker_exceptions.X2GoBrokerAgentException('Query to local X2Go Broker Agent failed with no response')

def call_remote_broker_agent(username, mode, cmdline_args=[], remote_agent=None):
    """\
    Launch remote X2Go Broker Agent via SSH and process its output.

    @param username: run the broker agent for this user
    @type username: C{unicode}
    @param mode: execution mode of the broker (listsessions, getservers, etc.)
    @type mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    if remote_agent is None:
        logger_error.error('With the SSH agent-query-mode remote agent (hostname, port) has to be specified!')

    cmd_line = [
        '{x2gobroker_agent_binary}'.format(x2gobroker_agent_binary=x2gobroker.defaults.X2GOBROKER_AGENT_CMD),
        '{username}'.format(username=username),
        '{mode}'.format(mode=mode),
    ]

    for cmdline_arg in cmdline_args:
        cmd_line.append('\'{arg}\''.format(arg=cmdline_arg))

    remote_username = x2gobroker.defaults.X2GOBROKER_AGENT_USER
    remote_hostname = remote_agent[u'hostname']
    if remote_agent.has_key(u'port'):
        remote_port = int(remote_agent[u'port'])
    else:
        remote_port = 22

    # now, connect and use paramiko Client to negotiate SSH2 across the connection
    try:
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.WarningPolicy())
        client.connect(remote_hostname, remote_port, remote_username, look_for_keys=True, allow_agent=True)

        result = []
        ssh_transport = client.get_transport()
        if ssh_transport and ssh_transport.is_authenticated():
            cmd = ' '.join(cmd_line)
            cmd = 'sh -c \"{cmd}\"'.format(cmd=cmd)
            logger_broker.debug('Executing agent command on remote host ({remote_agent}): {cmd}'.format(remote_agent=remote_agent['hostname'], cmd=cmd))
            (stdin, stdout, stderr) = client.exec_command(cmd)
            result = stdout.read().split('\n')
        client.close()
        if result and result[0].startswith('OK'):
            return [ r for r in result[1:] if r ]
    except (paramiko.SSHException, paramiko.AuthenticationException, paramiko.BadHostKeyException, socket.error):
        raise x2gobroker.x2gobroker_exceptions.X2GoBrokerAgentException('Query to remote X2Go Broker Agent (user: {user}, hostname: {hostname}, port: {port}) failed'.format(user=remote_username, hostname=remote_hostname, port=remote_port))


def ping(query_mode='LOCAL', remote_agent=None):
    """\
    Ping X2Go Broker Agent.

    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    username='foo'
    if unicode(query_mode).upper() == u'LOCAL':
        return call_local_broker_agent(username, mode='ping')
    else:
        return call_remote_broker_agent(username, mode='ping', remote_agent=remote_agent)


def list_sessions(username, query_mode='LOCAL', remote_agent=None):
    """\
    Query X2Go Broker Agent for a session list for a given username.

    @param username: run the query on behalf of this username
    @type username: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    if unicode(query_mode).upper() == u'LOCAL':
        return call_local_broker_agent(username, mode='listsessions')
    else:
        return call_remote_broker_agent(username, mode='listsessions', remote_agent=remote_agent)


def suspend_session(username, session_name, query_mode='LOCAL', remote_agent=None):
    """\
    Trigger a session suspensions via the X2Go Broker Agent.

    @param username: suspend the session on behalf of this username
    @type username: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    if unicode(query_mode).upper() == u'LOCAL':
        return call_local_broker_agent(username, mode='suspendsession', cmdline_args=[session_name, ], )
    else:
        return call_remote_broker_agent(username, mode='suspendsession', cmdline_args=[session_name, ], remote_agent=remote_agent)


def terminate_session(username, session_name, query_mode='LOCAL', remote_agent=None):
    """\
    Trigger a session termination via the X2Go Broker Agent.

    @param username: terminate the session on behalf of this username
    @type username: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    if unicode(query_mode).upper() == u'LOCAL':
        return call_local_broker_agent(username, mode='terminatesession', cmdline_args=[session_name, ], )
    else:
        return call_remote_broker_agent(username, mode='terminatesession', cmdline_args=[session_name, ], remote_agent=remote_agent)


def has_sessions(username, query_mode='LOCAL', remote_agent=None):
    """\
    Query X2Go Broker Agent to detect running/suspended sessions on
    the remote X2Go Server (farm).

    @param username: run the query on behalf of this username
    @type username: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    @return: (<has-running-sessions>, <has-suspended-session>), a tuple of two Boolean values
    @rtype: C{tuple}

    """
    _session_list = list_sessions(username, query_mode=query_mode, remote_agent=remote_agent)
    return ([ s.split('|')[3] for s in _session_list if s.split('|')[4] == 'R' ], [ s.split('|')[3] for s in _session_list if s.split('|')[4] == 'S' ])

def find_busy_servers(username, query_mode='LOCAL', remote_agent=None):
    """\
    Query X2Go Broker Agent for a list of  servers with running
    and/or suspended sessions and a percentage that tells about
    the busy-state of the server.

    The result is independent from the username given.

    @param username: run the query on behalf of this username
    @type username: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    if unicode(query_mode).upper() == u'LOCAL':
        server_list = call_local_broker_agent(username, mode='findbusyservers')
    else:
        server_list = call_remote_broker_agent(username, mode='findbusyservers', remote_agent=remote_agent)

    server_usage = {}

    if server_list:
        for server_item in server_list:
            if ':' in server_item:
                usage, server = server_item.split(':')
                server_usage.update({ server: int(usage) })

    return server_usage


def add_authorized_key(username, pubkey_hash, authorized_keys_file='%h/.x2go/authorized_keys', query_mode='LOCAL', remote_agent=None):
    """\
    Add a public key hash to the user's authorized_keys file.

    @param username: run the query on behalf of this username
    @type username: C{unicode}
    @param pubkey_hash: the public key hash as found in SSH authorized_keys files
    @type pubkey_hash: C{unicode}
    @param authorized_keys_file: the full path to the remote X2Go server's authorized_keys file
    @type authorized_keys_file: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    if unicode(query_mode).upper() == u'LOCAL':
        return call_local_broker_agent(username, mode='addauthkey', cmdline_args=[pubkey_hash, authorized_keys_file, ])
    else:
        return call_remote_broker_agent(username, mode='addauthkey', cmdline_args=[pubkey_hash, authorized_keys_file, ], remote_agent=remote_agent)


def delete_authorized_key(username, pubkey_hash, authorized_keys_file='%h/.x2go/authorized_keys', query_mode='LOCAL', remote_agent=None, delay_deletion=0):
    """\
    Remove a public key hash from the user's authorized_keys file.

    @param username: run the query on behalf of this username
    @type username: C{unicode}
    @param pubkey_hash: the public key hash as found in SSH authorized_keys files
    @type pubkey_hash: C{unicode}
    @param authorized_keys_file: the full path to the remote X2Go server's authorized_keys file
    @type authorized_keys_file: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    # this is for the logger output
    if remote_agent in  (None, u'LOCAL'):
        _hostname = 'LOCAL'
    else:
        _hostname = remote_agent['hostname']

    if delay_deletion > 0:
        delayed_execution(delete_authorized_key, delay=delay_deletion, username=username, pubkey_hash=pubkey_hash, authorized_keys_file=authorized_keys_file, query_mode=query_mode, remote_agent=remote_agent, )
        logger_broker.debug('Scheduled deletion of authorized key in {delay}s: user={user}, host={host}'.format(delay=delay_deletion, user=username, host=_hostname))
    else:
        if unicode(query_mode).upper() == u'LOCAL':
            return call_local_broker_agent(username, mode='delauthkey', cmdline_args=[pubkey_hash, authorized_keys_file, ])
        else:
            return call_remote_broker_agent(username, mode='delauthkey', cmdline_args=[pubkey_hash, authorized_keys_file, ], remote_agent=remote_agent)


def get_servers(username, query_mode='LOCAL', remote_agent=None):
    """\
    Query X2Go Broker Agent for the list of currently used servers.

    The result is independent from the username given.

    @param username: run the query on behalf of this username
    @type username: C{unicode}
    @param query_mode: query mode used when calling X2Go Broker Agent (C{LOCAL} or C{SSH})
    @type query_mode: C{unicode}
    @param remote_agent: information about the remote agent that is to be called.
    @type remote_agent: C{dict}

    """
    if unicode(query_mode).upper() == u'LOCAL':
        return call_local_broker_agent(username, mode='getservers')
    else:
        return call_local_broker_agent(username, mode='getservers', remote_agent=remote_agent)


def genkeypair(local_username, client_address, key_type='RSA'):
    """\
    Generate an SSH pub/priv key pair without writing the private key to file.

    @param local_username: the key is for this user
    @type local_username: C{unicode}
    @param client_address: the key is only valid for this client
    @type client_address: C{unicode}
    @param key_type: either of: RSA, DSA
    @type key_type: C{unicode}

    """
    key = None
    pubkey = None
    privkey = None

    # generate key pair
    if unicode(key_type) == u'RSA':
        key = paramiko.RSAKey.generate(2048)
    elif unicode(key_type) == u'DSA':
        key = paramiko.DSSKey.generate(1024)

    if key:

        # assemble the public key
        if key_type == "RSA":
            pubkey_type = 'ssh-rsa'
        elif key_type == "DSA":
            pubkey_type = 'ssh-dss'

        # FIXME: the from option does not work properly by some reason. Fix it later
        #pubkey = "from={client_address},no-X11-forwarding,no-pty,no-user-rc {pubkey_type} {pubkey} {local_username}@{client_address}".format(pubkey=key.get_base64(), pubkey_type=pubkey_type, local_username=local_username, client_address=client_address)
        pubkey = "no-X11-forwarding,no-pty,no-user-rc {pubkey_type} {pubkey} {local_username}@{client_address}".format(pubkey=key.get_base64(), pubkey_type=pubkey_type, local_username=local_username, client_address=client_address)

        # assemble the private key
        privkey_obj = cStringIO.StringIO()
        key.write_private_key(privkey_obj)
        privkey = privkey_obj.getvalue()

    return (pubkey, privkey)
