# -*- coding: utf-8 -*-

# Copyright (C) 2012-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
# Copyright (C) 2012-2013 by Oleksandr Shneyder <oleksandr.shneyder@obviously-nice.de>
#
# X2Go Session Broker is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# X2Go Session Broker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
L{base.X2GoBroker} class - base skeleton for X2GoBroker implementations

"""
__NAME__ = 'x2gobroker-pylib'

# modules
import types
import copy
import uuid
import netaddr
import random
import time

# X2Go Broker modules
import x2gobroker.config
import x2gobroker.defaults
import x2gobroker.agent

from x2gobroker.loggers import logger_broker, logger_error

class X2GoBroker(object):
    """\
    L{base.X2GoBroker} is an abstract class for X2Go broker implementations.

    This class needs to be inherited from a concrete broker class.

    Currently available broker classes are::
        L{zeroconf.X2GoBroker} (working)
        L{inifile.X2GoBroker} (in prep)
        L{ldap.X2GoBroker} (in prep)

    """

    backend_name = 'base'
    nameservice_module = None
    authmech_module = None

    def __init__(self, config_file=None, config_defaults=None):
        """\
        Initialize a new X2GoBroker instance to control X2Go session through an
        X2Go Client with an intermediate session broker.

        """
        self.config_file = config_file
        if self.config_file is None: self.config_file = x2gobroker.defaults.X2GOBROKER_CONFIG
        if config_defaults is None: config_defaults = x2gobroker.defaults.X2GOBROKER_CONFIG_DEFAULTS
        self.config = x2gobroker.config.X2GoBrokerConfigFile(config_files=self.config_file, defaults=config_defaults)
        self.enabled =  self.config.get_value(self.backend_name, 'enable')

        self._dynamic_cookie_map = {}
        self._client_address = None

    def __del__(self):
        """\
        Cleanup on destruction of an L{X2GoBroker} instance.

        """
        pass

    def is_enabled(self):
        """\
        Check if this backend has been enabled in the configuration file.

        """
        return self.enabled

    def enable(self):
        """\
        Enable this broker backend.

        """
        self.enabled = True

    def disable(self):
        """\
        Disable this broker backend.

        """
        self.enabled = False

    def set_client_address(self, address):
        """\
        Set the client IP address.

        @param ip: the client IP
        @type ip: C{unicode}

        """
        if netaddr.valid_ipv6(address):
            pass
        elif netaddr.valid_ipv4(address):
            pass
        else:
            self._client_address = None
            raise ValueError('address {address} is neither a valid IPv6 nor a valid IPv4 address'.format(address=address))
        self._client_address = netaddr.IPAddress(address)

    def get_client_address(self):
        """\
        Get the client IP address (if set).

        @return the client IP (either IPv4 or IPv6)
        @rtype ip: C{unicode}

        """
        if self._client_address:
            return unicode(self._client_address)
        else:
            return None

    def get_client_address_type(self):
        """\
        Get the client IP address type of the client address (if set).

        @return the client address type (4: IPv4, 6: IPv6)
        @rtype ip: C{int}

        """
        return self._client_address.version

    def get_global_config(self):
        """\
        Get the global section of the configuration file.

        @return: all global configuration parameters
        @rtype: C{dict}

        """
        return self.config.get_section('global')

    def get_global_value(self, option):
        """\
        Get the configuration setting for an option in the global section of the
        configuration file.

        @param option: option name in the global configuration section
        @type option: C{unicode}

        @return: the value for the given global C{option}
        @rtype: C{bool}, C{unicode}, C{int} or C{list}

        """
        return self.config.get_value('global', option)

    def get_backend_config(self):
        """\
        Get the configuration section of a specific backend.

        @return: all backend configuration parameters
        @rtype: C{dict}

        """
        return self.config.get_section(self.backend_name)

    def get_backend_value(self, backend='zeroconf', option='enable'):
        """\
        Get the configuration setting for backend C{backend} and option
        C{option}.

        @param backend: the name of the backend
        @type backend: C{unicode}
        @param option: option name of the backend's configuration section
        @type option: C{unicode}

        @return: the value for the given C{backend} C{option}
        @rtype: C{bool}, C{unicode}, C{int} or C{list}

        """
        return self.config.get_value(backend, option)

    def get_profile_ids(self):
        """\
        Retrieve the complete list of session profile IDs.

        @return: list of profile IDs
        @rtype: C{list}

        """
        return []

    def get_profile_defaults(self):
        """\
        Get the session profile defaults, i.e. profile options that all
        configured session profiles have in common.

        The defaults are hard-coded in L{x2gobroker.defaults} for class
        L{x2gobroker.base.X2GoBroker}.

        @return: a dictionary containing the session profile defaults
        @rtype: C{dict}

        """
        profile_defaults = copy.deepcopy(x2gobroker.defaults.X2GOBROKER_SESSIONPROFILE_DEFAULTS['DEFAULT'])
        for key in copy.deepcopy(profile_defaults):
            if key.startswith('acl-'):
                del profile_defaults[key]
        return profile_defaults

    def get_acl_defaults(self):
        """\
        Get the ACL defaults for session profiles. The defaults are hard-coded
        in L{x2gobroker.defaults} for class L{x2gobroker.base.X2GoBroker}.

        @return: a dictionary containing the ACL defaults for all session profiles
        @rtype: C{dict}

        """
        acl_defaults = copy.deepcopy(x2gobroker.defaults.X2GOBROKER_SESSIONPROFILE_DEFAULTS['DEFAULT'])
        for key in copy.deepcopy(acl_defaults):
            if not key.startswith('acl-'):
                del acl_defaults[key]
        return acl_defaults

    def get_profile(self, profile_id):
        """\
        Get the session profile for profile ID <profile_id>.

        @param profile_id: the ID of a profile
        @type profile_id: C{unicode}

        @return: a dictionary representing the session profile for ID <profile_id>
        @rtype: C{dict}

        """
        return {}

    def get_profile_broker(self, profile_id):
        """\
        Get broker-specific session profile options from the session profile with profile ID <profile_id>.

        @param profile_id: the ID of a profile
        @type profile_id: C{unicode}

        @return: a dictionary representing the session profile for ID <profile_id>
        @rtype: C{dict}

        """
        return {}

    def get_profile_acls(self, profile_id):
        """\
        Get the ACLs for session profile with profile ID <profile_id>.

        @param profile_id: the ID of a profile
        @type profile_id: C{unicode}

        @return: a dictionary representing the ACLs for session profile with ID <profile_id>
        @rtype: C{dict}

        """
        return {}

    def check_profile_acls(self, username, acls):
        """\
        Test if a given user can get through an ACL check using <acls> as a list
        of allow and deny rules.

        @param username: the username of interest
        @type username: C{unicode}
        @param acls: a dictionary data structure containing ACL information (see L{defaults.X2GOBROKER_SESSIONPROFILE_DEFAULTS})
        @type acls: C{dict}

        """
        ### extract ACLs evaluation orders

        _acls = self.get_acl_defaults()
        _acls.update(acls)
        _order = {}
        _order[u'users'] = _order[u'groups'] = _order[u'clients'] = _acls[u'acl-any-order']

        try: _order[u'users'] = _acls[u'acl-users-order']
        except KeyError: pass
        try: _order[u'groups'] = _acls[u'acl-groups-order']
        except KeyError: pass
        try: _order[u'clients'] = _acls[u'acl-clients-order']
        except KeyError: pass

        # to pass an ACL test, all three keys in the dict below have to be set to True
        # if one stays False, the related session profile will not be returned to the querying
        # X2Go client...
        _grant_availability = {
            'by_user': False,
            # FIXME: leaving the group access to False for now, we need methods that give us a generic
            # acces to the list of groups a user belongs to. One generic access is asking libnss, but:
            # are there others?
            'by_group': False,
            # FIXME: set the client access to True for now as we have not a way to check that available...
            'by_client': True,
        }

        ### CHECKING on a per-client basis...

        ### clients access is granted first, if that fails then we return False here...

        if len( _acls[u'acl-clients-allow'] + _acls[u'acl-clients-deny'] ) > 0:

            _acls_clients_allow = copy.deepcopy(_acls[u'acl-clients-allow'])
            _acls_clients_deny = copy.deepcopy(_acls[u'acl-clients-deny'])

            _allow_client = False
            _deny_client = False

            for idx, item in enumerate(_acls_clients_allow):
                if unicode(item) == u'ALL':
                    _acls_clients_allow[idx] = '0.0.0.0/0'
                    _acls_clients_allow.insert(idx, '::/0')
            for idx, item in enumerate(_acls_clients_deny):
                if unicode(item) == u'ALL':
                    _acls_clients_deny[idx] = '0.0.0.0/0'
                    _acls_clients_deny.insert(idx, '::/0')

            _allow_address_set = []
            _deny_address_set = ['ALL']
            try:
                _allow_address_set = netaddr.IPSet(_acls_clients_allow)
                _deny_address_set = netaddr.IPSet(_acls_clients_deny)
            except netaddr.core.AddrFormatError, e:
                logger_error.error('base_broker.X2GoBroker.check_acls(): netaddr.core.AddrFormatError - {why}'.format(why=str(e)))
            except ValueError, e:
                logger_error.error('base_broker.X2GoBroker.check_acls(): ValueError - {why}'.format(why=str(e)))

            _allow_client = self._client_address in _allow_address_set
            _deny_client = self._client_address in _deny_address_set

            if not (_allow_client or _deny_client):
                # client was not in either of the rules, so we presume that the client is allowed to access
                _grant_availability[u'by_client']
            if _order[u'clients'] == 'allow-deny':
                _grant_availability[u'by_client'] = _allow_client and (not _deny_client)
            else:
                _grant_availability[u'by_client'] = (not _deny_client) or _allow_client

            if not _grant_availability[u'by_client']:
                return False

        ### no user/group ACLs are in use, allow access then...

        if len(_acls[u'acl-users-allow'] + _acls[u'acl-users-deny'] + _acls[u'acl-groups-allow'] + _acls[u'acl-groups-deny']) == 0:
            return True

        ### CHECKING on a per-user basis...

        _allow_user_override = False
        _explicitly_deny_user = False
        if len( _acls[u'acl-users-allow'] + _acls[u'acl-users-deny'] ) > 0:

            _allow_user = False
            _deny_user = False

            if username in _acls[u'acl-users-allow'] or u'ALL' in _acls[u'acl-users-allow']:
                _allow_user_override = True
                _allow_user = True

            if username in _acls[u'acl-users-deny']:
                _explicitly_deny_user = True
            if _explicitly_deny_user or (u'ALL' in _acls[u'acl-users-deny']):
                _deny_user = True

            if _order[u'users'] == 'allow-deny':
                _grant_availability[u'by_user'] = (_allow_user or _deny_user) and (_allow_user and (not _deny_user))
            else:
                _grant_availability[u'by_user'] = (_allow_user or _deny_user) and ((not _deny_user) or _allow_user)

            # if a user has been granted access directly, then the corresponding session profile(s)
            # will be provided to him/her, it does not matter what the group acl will have to say to this...
            if _grant_availability[u'by_user']:
                return True

        ### CHECKING on a per-group basis...

        if len(_acls[u'acl-groups-allow'] + _acls[u'acl-groups-deny']) > 0:
            _allow_group = False
            _deny_group = False

            _user_groups = [u'ALL'] + self.get_user_groups(username, primary_groups=not self.get_global_value('ignore-primary-group-memberships'))

            _allow_group = bool(len(set(_user_groups).intersection( set(_acls[u'acl-groups-allow']) )))
            _deny_group = bool(len(set(_user_groups).intersection( set(_acls[u'acl-groups-deny']) )))

            if _order[u'groups'] == 'allow-deny':
                _grant_availability[u'by_group'] = (_allow_group or _deny_group) and (_allow_group and (not _deny_group))
            else:
                _grant_availability[u'by_group'] = (_allow_group or _deny_group) and (not _deny_group) or _allow_group

            # if a group has been granted access, with one exception: if the thread model for users is
            # allow-deny, then we presume that the acl-users-deny entry has precendence over
            # acl-groups-allow/acl-groups-deny.
            if (_grant_availability[u'by_group'] and not _explicitly_deny_user) or _allow_user_override:
                return True

        return False

    def test_connection(self):
        #if($cgi->param('task') eq 'testcon')
        #{
        #   for ( my $i=0;$i<2*1024*1024;$i++ )
        #   {
        #           print int(rand(9));
        #   }
        #   print $cgi->end_html();
        #   exit (0);
        #}
        return 'OK'

    def _import_authmech_module(self, mech='pam'):
        try:
            if self.authmech_module is None:
                exec("import x2gobroker.authmechs.{mech}_authmech as _authmech_module".format(mech=mech))
                self.authmech_module = _authmech_module
            return True
        except ImportError:
            return False

    def _do_authenticate(self, username='', password=''):

        if self._import_authmech_module(mech=self.get_authentication_mechanism()):
            logger_broker.debug('base_broker.X2GoBroker._do_authenticate(): authenticating user={username} with password=<hidden> against backend={backend}.'.format(username=username, backend=self.backend_name))
            return self.authmech_module.X2GoBrokerAuthMech().authenticate(username, password)
        else:
            return False

    def get_authentication_mechanism(self):
        """\
        Get the name of the authentication mechanism that is configured for this
        X2Go Session Broker instance.

        @return: auth-mech name
        @rtype: C{unicode}

        """
        _default_auth_mech = "pam"
        _auth_mech = ""

        if self.config.has_value(self.backend_name, 'auth-mech') and self.config.get_value(self.backend_name, 'auth-mech'):
            _auth_mech = self.config.get_value(self.backend_name, 'auth-mech').lower()
            logger_broker.debug('base_broker.X2GoBroker.get_authentication_mechanism(): found auth-mech in backend config section »{backend}«: {value}. This one has precendence over the default value.'.format(backend=self.backend_name, value=_auth_mech))

        elif self.config.has_value('global', 'default-auth-mech'):
            _default_auth_mech = self.config.get_value('global', 'default-auth-mech').lower()
            logger_broker.debug('base_broker.X2GoBroker.get_authentication_mechanism(): found default-auth-mech in global config section: {value}'.format(value=_default_auth_mech))

        return unicode(_auth_mech) or unicode(_default_auth_mech)

    def get_agent_query_mode(self):
        """\
        Get the agent query mode (LOCAL or SSH, normally) that is configured for this
        X2Go Session Broker instance.

        @return: agent query mode
        @rtype: C{unicode}

        """
        _default_agent_query_mode = "LOCAL"
        _agent_query_mode = ""
        if self.config.has_value(self.backend_name, 'agent-query-mode') and self.config.get_value(self.backend_name, 'agent-query-mode'):
            _agent_query_mode = self.config.get_value(self.backend_name, 'agent-query-mode').lower()
            logger_broker.debug('base_broker.X2GoBroker.get_agent_query_mode(): found agent-query-mode in backend config section »{backend}«: {value}. This one has precendence over the default value.'.format(backend=self.backend_name, value=_agent_query_mode))

        elif self.config.has_value('global', 'default-agent-query-mode'):
            _default_agent_query_mode = self.config.get_value('global', 'default-agent-query-mode').lower()
            logger_broker.debug('base_broker.X2GoBroker.get_agent_query_mode(): found default-agent-query-mode in global config section: {value}'.format(value=_default_agent_query_mode))

        return unicode(_agent_query_mode) or unicode(_default_agent_query_mode)

    def use_session_autologin(self, profile_id):
        """\
        Detect if the given profile is configured to try automatic session
        logons.

        @return: C{True} to denote that automatic session login should be attempted
        @rtype: C{bool}

        """
        _default_session_autologin = False
        _session_autologin = False
        _profile = self.get_profile_broker(profile_id)
        if _profile and _profile.has_key(u'broker-session-autologin') and _profile['broker-session-autologin']:
            _session_autologin = _profile[u'broker-session-autologin']
            logger_broker.debug('base_broker.X2GoBroker.get_session_autologin(): found broker-session-autologin in session profile with ID {id}: {value}. This one has precendence over the default value.'.format(id=profile_id, value=_session_autologin))

        elif self.config.has_value('global', 'default-session-autologin'):
            _default_session_autologin = self.config.get_value('global', 'default-session-autologin')
            logger_broker.debug('base_broker.X2GoBroker.get_session_autologin(): found default-session-autologin in global config section: {value}'.format(value=_default_session_autologin))

        return _session_autologin or _default_session_autologin

    def get_authorized_keys_file(self, profile_id):
        """\
        Get the default location of server-side authorized_keys files used with
        the X2Go Session Broker.

        The file location can be configured broker-wide. It is also possible to
        provide a broker-authorized-keys file in session profiles. The latter
        will override the broker-wide conigured file location.

        @return: authorized_keys location on the remote server
        @rtype: C{unicode}

        """
        _default_authorized_keys_file = "%h/.x2go/authorized_keys"
        _authorized_keys_file = ""
        _profile = self.get_profile_broker(profile_id)
        if _profile and _profile.has_key(u'broker-authorized-keys') and _profile['broker-authorized-keys']:
            _authorized_keys_file = _profile[u'broker-authorized-keys']
            logger_broker.debug('base_broker.X2GoBroker.get_authorized_keys_file(): found broker-authorized-keys in session profile with ID {id}: {value}. This one has precendence over the default value.'.format(id=profile_id, value=_authorized_keys_file))

        elif self.config.has_value('global', 'default-authorized-keys'):
            _default_authorized_keys_file = self.config.get_value('global', 'default-authorized-keys').lower()
            logger_broker.debug('base_broker.X2GoBroker.get_authorized_keys_file(): found default-authorized-keys in global config section: {value}'.format(value=_default_authorized_keys_file))

        return unicode(_authorized_keys_file) or unicode(_default_authorized_keys_file)

    def get_userdb_service(self):
        """\
        Get the name of the backend being used for retrieving user information from the
        system.

        @return: user service name
        @rtype: C{unicode}

        """
        _user_db = "libnss"
        if self.config.has_value('global', 'default-user-db'):
            _user_db = self.config.get_value('global', 'default-user-db').lower() or _user_db

        if self.config.has_value(self.backend_name, 'user-db'):
            _user_db = self.config.get_value(self.backend_name, 'user-db').lower() or _user_db

        return unicode(_user_db)

    def get_groupdb_service(self):
        """\
        Get the name of the backend being used for retrieving group information from the
        system.

        @return: group service name
        @rtype: C{unicode}

        """
        _group_db = "libnss"
        if self.config.has_value('global', 'default-group-db'):
            _group_db = self.config.get_value('global', 'default-group-db').lower() or _group_db

        if self.config.has_value(self.backend_name, 'group-db'):
            _group_db = self.config.get_value(self.backend_name, 'group-db').lower() or _group_db

        return unicode(_group_db)

    def _import_nameservice_module(self, service='libnss'):
        try:
            if self.nameservice_module is None:
                exec("import x2gobroker.nameservices.{service}_nameservice as _nameservice_module".format(service=service))
                self.nameservice_module = _nameservice_module
            return True
        except ImportError:
            return False

    def has_user(self, username):
        """\
        Test if the broker knows user C{<username>}.

        @param username: test for existence of this user
        @type username: C{unicode}

        @return: returns C{True} if a user exists
        @rtype: C{bool}

        """
        if self._import_nameservice_module(service=self.get_userdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().has_user(username=username)
        else:
            return False

    def get_users(self):
        """\
        Get list of known users.

        @return: returns list of known users
        @rtype: C{list}

        """
        if self._import_nameservice_module(service=self.get_userdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().get_users()
        else:
            return False

    def has_group(self, group):
        """\
        Test if the broker knows group C{<group>}.

        @param group: test for existence of this group
        @type group: C{unicode}

        @return: returns C{True} if a group exists
        @rtype: C{bool}

        """
        if self._import_nameservice_module(service=self.get_groupdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().has_group(group=group)
        else:
            return False

    def get_groups(self):
        """\
        Get list of known groups.

        @return: returns list of known groups
        @rtype: C{list}

        """
        if self._import_nameservice_module(service=self.get_groupdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().get_groups()
        else:
            return False

    def get_primary_group(self, username):
        """\
        Get the primary group of a given user.

        @param username: get primary group for this username
        @type username: C{unicode}

        @return: returns the name of the primary group
        @rtype: C{unicode}

        """
        if self._import_nameservice_module(service=self.get_groupdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().get_primary_group(username)
        else:
            return False

    def is_group_member(self, username, group, primary_groups=False):
        """\
        Check if a user is member of a given group.

        @param username: check group membership of this user
        @type username: C{unicode}
        @param group: test if user is member of this group
        @type group: C{unicode}
        @param primary_groups:  if C{True}, test for primary group membership, as well
        @type primary_groups: C{bool}

        @return: returns C{True} if the user is member of the given group
        @rtype: C{bool}

        """
        if self._import_nameservice_module(service=self.get_groupdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().is_group_member(username=username, group=group, primary_groups=primary_groups)
        else:
            return []

    def get_group_members(self, group, primary_groups=False):
        """\
        Get the list of members in group C{<group>}.

        @param group: valid group name
        @type group: C{unicode}
        @param primary_groups: include primary groups found with the user db service
        @type primary_groups: C{bool}

        @return: list of users belonging to the given group
        @rtype: C{list}

        """
        if self._import_nameservice_module(service=self.get_groupdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().get_group_members(group=group, primary_groups=primary_groups)
        else:
            return []

    def get_user_groups(self, username, primary_groups=False):
        """\
        Get all groups a given user is member of.

        @param username: get groups for this user
        @type username: C{unicode}
        @param primary_groups:  if C{True}, include the user's primary group in the group list
        @type primary_groups: C{bool}

        @return: list of groups the given user is member of
        @rtype: C{list}

        """
        if self._import_nameservice_module(service=self.get_groupdb_service()):
            return self.nameservice_module.X2GoBrokerNameService().get_user_groups(username=username, primary_groups=primary_groups)
        else:
            return []

    def check_access(self, username='', password='', cookie=None, ):
        """\
        Check if a given user with a given password may gain access to the
        X2Go session broker.

        @param username: a username known to the session broker
        @type username: C{unicode}
        @param password: a password that authenticates the user against the X2Go session broker
        @type password: C{unicode}
        @param cookie: an extra (static or dynamic) authentication token
        @type cookie: C{unicode}

        @return: returns C{True} if the authentication has been successful
        @rtype: C{bool}

        """
        ### FOR INTRANET LOAD BALANCER WE MAY JUST ALLOW ACCESS TO EVERYONE
        ### This is handled through the config file, normally /etc/x2go/x2gobroker.conf

        if not self.config.get_value('global', 'check-credentials'):
            logger_broker.debug('base_broker.X2GoBroker.check_access(): access is granted without checking credentials, prevent this in {configfile}'.format(configfile=self.config_file))
            return True

        ### IMPLEMENT YOUR AUTHENTICATION LOGIC IN THE self._do_authenticate(**kwargs) METHOD
        ### when inheriting from the base.X2GoBroker class.

        access = False
        access = self._do_authenticate(username=username, password=password)
        logger_broker.debug('base_broker.X2GoBroker.check_access(): result of authentication check is: {access}'.format(access=access))

        ### HANDLING OF DYNAMIC AUTHENTICATION ID HASHES

        # using cookie authentication as extra security?
        if self.config.get_value('global', 'require-cookie-authentication'):

            if type(cookie) is types.StringType:
                cookie = unicode(cookie)

            if self.config.get_value('global', 'use-static-cookie'):

                # evaluate access based on static authentication ID feature
                access = access and ( cookie == self.config.get_value('global', 'my-cookie') )

            else:

                # evaluate access based on dynamic authentication ID feature
                if self._dynamic_cookie_map.has_key(username):
                    access = access and ( cookie == self._dynamic_cookie_map[username] )
                    if access:
                        self._dynamic_cookie_map[username] = uuid.uuid5(namespace=cookie, name=username)

                else:
                    access = access and ( cookie == self.config.get_value('global', 'my-cookie') )
                    if access:
                        # generate a first uuid, initialize the dynamic authencation ID security feature
                        self._dynamic_cookie_map[username] = uuid.uuid4()

        return access

    def get_next_cookie(self, username):
        """\
        Get the next expected authentication cookie for the given user name.

        @param username: query next authentication cookie for this user
        @type username: C{unicode}

        @return: returns next authentication cookie for the given username, None if no cookie has been generated, yet
        @rtype: C{unicode} or C{None}

        """
        try:
            return self._dynamic_cookie_map[username]
        except KeyError:
            return None


    def get_remote_agent(self, profile_id, exclude_agents=[], ):
        """\
        Randomly choose a remote agent for agent query.

        @param profile_id: choose remote agent for this profile ID
        @type profile_id: C{unicode}
        @param exclude_agents: a list of remote agent dict objects to be exclude from the random choice
        @type exclude_agents: C{list}

        """
        remote_agent = None

        agent_query_mode = self.get_agent_query_mode().upper()
        if agent_query_mode == u'SSH':

            profile = self.get_profile(profile_id)
            server_list = profile[u'host']
            random.shuffle(server_list)

            for agent in exclude_agents:
                if agent['hostname'] in server_list:
                    server_list.remove(agent['hostname'])

            while server_list:

                remote_agent_server = server_list[-1]
                remote_agent_port = profile[u'sshport']
                remote_agent = {u'hostname': remote_agent_server, u'port': remote_agent_port, }

                if x2gobroker.agent.ping(query_mode=agent_query_mode, remote_agent=remote_agent):
                    break

                server_list = server_list[0:-1]

            else:
                logger_broker.warning('base_broker.X2GoBroker.get_remote_agent(): failed to allocate any broker agent (query-mode: {query_mode}, remote_agent: {remote_agent})'.format(query_mode=agent_query_mode, remote_agent=remote_agent))

        elif agent_query_mode == u'LOCAL':
            # use a non-False value here, not used anywhere else...
            remote_agent = 'LOCAL'

        return remote_agent

    def list_profiles(self, username):
        """\
        Retrieve a list of available session profiles for the authenticated user.

        @param username: query session list for this user
        @type username: C{unicode}

        return: list of profile dictionaries
        rtype: C{dict}

        """
        list_of_profiles = {}
        for profile_id in self.get_profile_ids():
            profile = self.get_profile(profile_id)

            for key in profile.keys():
                if key.startswith('host='):
                    del profile[key]

            if self.use_session_autologin(profile_id):
                profile['autologin'] = True
                profile['key'] = '<will-be-provided-later>'

            # make sure that desktop sessions (that we know by name) do run with rootless=false
            if profile['command'] in x2gobroker.defaults.X2GO_DESKTOP_SESSIONS:
                profile['rootless'] = False

            acls = self.get_profile_acls(profile_id)

            if self.check_profile_acls(username, acls):

                remote_agent = self.get_remote_agent(profile_id)
                agent_query_mode = ( remote_agent == u'LOCAL') and u'LOCAL' or u'SSH'
                if remote_agent:
                    running_sessions, suspended_sessions = x2gobroker.agent.has_sessions(username, query_mode=agent_query_mode, remote_agent=remote_agent)

                    if set(profile['host']).intersection(set(running_sessions)): profile['status'] = u'R'
                    if set(profile['host']).intersection(set(suspended_sessions)): profile['status'] = u'S'

                profile['host'] = [profile['host'][0]]
                list_of_profiles.update({profile_id: profile, })

        return list_of_profiles

    def select_session(self, profile_id, username=None):
        """\
        Start/resume a session by selecting a profile name offered by the X2Go client.

        The X2Go server that the session is launched on is selected automatically by the X2Go session
        broker.

        @param profile_id: the selected profile ID. This matches one of the dictionary keys offered by the C{list_profiles} method
        @type profile_id: C{unicode}
        @param username: specify username that this operation runs for
        @type username: C{unicode}

        """
        profile = self.get_profile(profile_id)

        # if we have more than one server, pick one server randomly for X2Go Broker Agent queries
        server_list = profile[u'host']
        remote_agent = self.get_remote_agent(profile_id)

        if remote_agent:
            agent_query_mode = ( remote_agent == u'LOCAL') and u'LOCAL' or u'SSH'

        # detect best X2Go server for this user if load balancing is configured
        if remote_agent and len(server_list) >= 2 and username:

            # query remote agent for session info
            busy_servers = None
            try:
                busy_servers = x2gobroker.agent.find_busy_servers(username=username, query_mode=agent_query_mode, remote_agent=remote_agent)
            except x2gobroker.x2gobroker_exceptions.X2GoBrokerAgentException:
                pass

            if busy_servers is not None:

                # if we do not get here, we failed to query a valid agent...

                # when detecting the server load we have to support handling of differing subdomains (config
                # file vs. server load returned by x2gobroker agent). Best approach: all members of a multi-node
                # server farm either (a) do not have a subdomain in their hostname or (b) the subdomain is
                # is identical.

                # Example:
                #
                #    ts01, ts02 - hostnames as returned by agent
                #    ts01.intern, ts02.intern - hostnames configured in session profile option ,,host''
                #    -> this will result in the subdomain .intern being stripped off from the hostnames before detecting
                #       the best server for this user

                ### NORMALIZE (=reduce to hostname only) X2Go server names (as found in config) if possible
                server_list_normalized, subdomains_config = x2gobroker.utils.normalize_hostnames(server_list)

                ### NORMALIZE X2Go server names (as returned by x2gobroker agent)--only if the hostnames in the config share the same subdomain
                if len(subdomains_config) == 1:

                    busy_servers_normalized, subdomains_agent = x2gobroker.utils.normalize_hostnames(busy_servers)
                    if len(subdomains_agent) <= 1:

                        # all X2Go servers in the multi-node server farm are in the same DNS subdomain
                        # we can operate on hostname-only hostnames
                        server_list = server_list_normalized
                        busy_servers = busy_servers_normalized

                for server in server_list:
                    if server not in busy_servers.keys():
                        busy_servers[server] = 0

                busy_server_list = [ (load, server) for server, load in busy_servers.items() ]
                busy_server_list.sort()

                logger_broker.debug('base_broker.X2GoBroker.select_session(): load balancer analysis: {server_load}'.format(server_load=unicode(busy_server_list)))

                best_server = busy_server_list[0][1]

            else:
                logger_broker.warning('base_broker.X2GoBroker.select_session(): no broker agent could be contacted, this does not look good. We tried these agent hosts: {agent_hosts}'.format(agent_hosts=unicode(server_list)))
                if server_list: best_server = server_list[0]
                else: return { 'server': 'no-server-available', 'port': profile[u'sshport'], }

        else:
            if server_list: best_server = server_list[0]
            else: return { 'server': 'no-server-available', 'port': profile[u'sshport'], }

        # if we have an explicit IP address for best_server, let's use that instead...
        try:
            best_server = profile['host={hostname}'.format(hostname=best_server)]
        except KeyError:
            pass

        selected_session = {
            'server': best_server,
            'port': profile[u'sshport'],
        }

        # find already running/suspended sessions and resume the first one found
        if remote_agent and server_list and username:

            session_list = x2gobroker.agent.list_sessions(username=username, query_mode=agent_query_mode, remote_agent=remote_agent)
            if session_list:

                # if resuming, always select the first session in the list, there should only be one suspended session
                try:
                    running_sessions = []
                    suspended_sessions = []
                    for session_info in session_list:
                        if session_info.split('|')[4] == 'R':
                            running_sessions.append(session_info)
                        if session_info.split('|')[4] == 'S':
                            suspended_sessions.append(session_info)

                    # we prefer suspended sessions for resuming if we find sessions with both states of activity
                    if suspended_sessions:
                        session_info = suspended_sessions[0]
                    else:
                        session_info = running_sessions[0]
                        x2gobroker.agent.suspend_session(username=username, session_name=session_info.split('|')[1], query_mode=agent_query_mode)
                        # this is the turn-around in x2gocleansessions, so waiting as along as the daemon that will suspend the session
                        time.sleep(2)
                        session_info = session_info.replace('|R|', '|S|')
                    server_name = session_info.split('|')[3]

                    # if we have an explicit IP address for server_name, let's use that instead...
                    try:
                        server_name = profile['host={hostname}'.format(hostname=server_name)]
                    except KeyError:
                        pass

                    selected_session.update({
                        'server': server_name,
                        'session_info': session_info,
                    })

                except IndexError:

                    # FIXME: if we get here, we have to deal with a broker session info entry in the X2Go session database
                    pass


        # session autologin feature
        if remote_agent and self.use_session_autologin(profile_id) and username:

            # FIXME: we somehow have to find out about the username of the person at the broker client-side...
            # using the username used for broker login for now...
            pubkey, privkey = x2gobroker.agent.genkeypair(local_username=username, client_address=self.get_client_address())
            x2gobroker.agent.add_authorized_key(username=username,
                                                pubkey_hash=pubkey,
                                                authorized_keys_file=self.get_authorized_keys_file(profile_id),
                                                query_mode=agent_query_mode,
                                                remote_agent=remote_agent,
            ),
            selected_session.update({
                'authentication_privkey': privkey,
            })
            x2gobroker.agent.delete_authorized_key(username=username,
                                                   pubkey_hash=pubkey,
                                                   authorized_keys_file=self.get_authorized_keys_file(profile_id),
                                                   query_mode=agent_query_mode,
                                                   remote_agent=remote_agent,
                                                   delay_deletion=20,
            )

        return selected_session

    def change_password(self, new='', old=''):
        """\
        Modify the authenticated user's password on the X2Go infrastructure (normally, one user
        in one X2Go site setup should have the same password on all machines).

        """
        return False
